#!python
# Copyright (c) 2000-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

# A Service for providing Komodo-specific directory/filename information.

import os
from os.path import basename, dirname, join, isfile, normpath
import sys
import re
import logging

from xpcom import components

log = logging.getLogger("koDirs")


#---- component implementation

class KoDirs:
    _com_interfaces_ = [components.interfaces.koIDirs]
    _reg_clsid_ = "{98435d6d-e24d-4057-a72a-dfcd1d282fae}"
    _reg_contractid_ = "@activestate.com/koDirs;1"
    _reg_desc_ = "Komodo Directory Information Service"

    _appdatadir_name = "komodoedit" # e.g. "KomodoIDE"

    def __init__(self):
        pass

    def _userAppDataPath(self):
        # Allow a special environment variable to override the User Data
        # Dir for a Komodo run. The main motivation for this is bug
        # 32270.
        envPath = os.environ.get("KOMODO_USERDATADIR")
        if envPath:
            path = os.path.expanduser(envPath)
        elif sys.platform.startswith("win"):
            # try to make this a unicode path bug #26404
            # SHGetFolderPath does not return unicode strings when
            # there is unicode data in the path
            from win32com.shell import shellcon, shell
            path = shell.SHGetFolderPath(0, shellcon.CSIDL_APPDATA, 0, 0)
            try:
                path = unicode(path)
            except:
                pass
            path = os.path.join(path, "ActiveState", self._appdatadir_name)
        elif sys.platform == 'darwin':
            if os.uname()[-1] == 'i386':
                #XXX Folder.FSFindFolder() fails with error -43 on x86. See 42669.
                path = os.path.expanduser('~/Library/Application Support/'
                                          + self._appdatadir_name)
            else:
                from Carbon import Folder, Folders
                path = Folder.FSFindFolder(Folders.kUserDomain,
                                           Folders.kApplicationSupportFolderType,
                                           Folders.kDontCreateFolder)
                path = os.path.join(path.FSRefMakePath(),
                                    self._appdatadir_name)
        else:
            path = os.path.expanduser("~/." + self._appdatadir_name)
        return path

    def _commonAppDataPath(self):
        if sys.platform.startswith("win"):
            # try to make this a unicode path bug #26404
            # SHGetFolderPath does not return unicode strings when
            # there is unicode data in the path
            from win32com.shell import shellcon, shell
            shellcon.CSIDL_COMMON_APPDATA = 0x23 # missing from shellcon
            path = shell.SHGetFolderPath(0, shellcon.CSIDL_COMMON_APPDATA, 0, 0)
            try:
                path = unicode(path)
            except:
                pass
            path = os.path.join(path, "ActiveState", self._appdatadir_name)
        elif sys.platform == 'darwin':
            if os.uname()[-1] == 'i386':
                #XXX Folder.FSFindFolder() fails with error -43 on x86. See 42669.
                path = '/Library/Application Support/' \
                       + self._appdatadir_name
            else:
                from Carbon import Folder, Folders
                path = Folder.FSFindFolder(Folders.kLocalDomain,
                                           Folders.kApplicationSupportFolderType,
                                           Folders.kDontCreateFolder)
                path = os.path.join(path.FSRefMakePath(),
                                    self._appdatadir_name)
        else:
            path = "/etc/" + self._appdatadir_name
        return path

    def _GetUserDataDirForVersion(self, major, minor):
        return os.path.join(self._userAppDataPath(),
                            "%s.%s" % (major, minor))

    def _GetCommonDataDirForVersion(self, major, minor):
        return os.path.join(self._commonAppDataPath(),
                            "%s.%s" % (major, minor))

    def get_userDataDir(self):
        infoSvc = components.classes["@activestate.com/koInfoService;1"].\
            getService(components.interfaces.koIInfoService)
        major, minor = infoSvc.version.split('.', 2)[:2]
        return self._GetUserDataDirForVersion(major, minor)

    def get_hostUserDataDir(self):
        # KOMODO_HOSTNAME is set by the koStart.c startup stuff.
        if os.environ.has_key("KOMODO_HOSTNAME"):
            hostname = os.environ["KOMODO_HOSTNAME"]
        else:
            import socket
            hostname = socket.gethostname()
            #XXX Should this be a warning?
        return os.path.join(self.get_userDataDir(), "host-"+hostname)

    def get_factoryCommonDataDir(self):
        #XXX Would be better called systemCommonDataDir, or
        #    defaultSystemCommonDataDir.
        infoSvc = components.classes["@activestate.com/koInfoService;1"].\
               getService(components.interfaces.koIInfoService)
        major, minor = infoSvc.version.split('.', 2)[:2]
        return self._GetCommonDataDirForVersion(major, minor)
        
    # XXX this code is duplicated in koPrefs.py since the global prefs
    # service needs the commonDataDir, but cannot access it via
    # koIDirs due to chicken-egg issue.  If you change this, change
    # koPrefs also!
    def get_commonDataDir(self):
        globalPrefs = components.classes["@activestate.com/koPrefService;1"]\
                      .getService(components.interfaces.koIPrefService).prefs
        method = "default"
        if globalPrefs.hasStringPref("commonDataDirMethod"):
            method = globalPrefs.getStringPref("commonDataDirMethod")
        if method not in ("default", "custom"):
            log.error("bogus Common Data Dir determination method, '%s', "
                      "falling back to default", method)
        if method == "default":
            return self.get_factoryCommonDataDir()
        elif method == "custom":
            return globalPrefs.getStringPref("customCommonDataDir")

    def get_mozBinDir(self):
        # By definition this is the directory of the main komodo/mozilla
        # executable. 'koDirs.py' is in the components directory, always one
        # dir under the moz bin dir.
        return dirname(dirname(__file__))

    __isDevTreeCache = None
    def _isDevTree(self):
        """Return true if this Komodo is running in a dev tree layout."""
        if self.__isDevTreeCache is None:
            landmark = os.path.join(self.get_mozBinDir(), "is_dev_tree.txt")
            self.__isDevTreeCache = os.path.isfile(landmark)
        return self.__isDevTreeCache

    def _getKomodoBitsDir(self):
        """Return the path to the "komodo-bits" dir in a dev tree.
        
        Note: this only makes sense if _isDevTree() is true.
        """
        if sys.platform == "darwin":
            # mozBinDir:     $mozSrc/mozilla/dist/Komodo.app/Contents/MacOS
            # komodoBitsDir: $mozSrc/mozilla/dist/komodo-bits
            komodoBitsDir = normpath(join(self.get_mozBinDir(),
                                     os.pardir, os.pardir, os.pardir,
                                     "komodo-bits"))
        else:
            # mozBinDir:     $mozSrc/mozilla/dist/bin
            # komodoBitsDir: $mozSrc/mozilla/dist/komodo-bits
            komodoBitsDir = join(dirname(self.get_mozBinDir()),
                                 "komodo-bits")
        return komodoBitsDir

    def get_supportDir(self):
        if self._isDevTree(): # in a development tree
            supportDir = join(self._getKomodoBitsDir(), "support")
        else:
            if sys.platform == "darwin":
                # mozBinDir:  /Applications/Komodo.app/Contents/MacOS
                # supportDir: /Applications/Komodo.app/Contents/SharedSupport
                supportDir = join(dirname(self.get_mozBinDir()),
                                  "SharedSupport")
            else:
                # mozBinDir:  <installdir>/lib/mozilla
                # supportDir: <installdir>/lib/support
                supportDir = join(dirname(self.get_mozBinDir()), "support")
        return supportDir

    def get_sdkDir(self):
        if self._isDevTree(): # in a development tree
            sdkDir = join(self._getKomodoBitsDir(), "sdk")
        else:
            if sys.platform == "darwin":
                sdkDir = join(self.get_supportDir(), "sdk")
            else:
                # mozBinDir: <installdir>/lib/mozilla
                # sdkDir:    <installdir>/lib/sdk
                sdkDir = join(dirname(self.get_mozBinDir()), "sdk")
        return sdkDir

    def get_docDir(self):
        if self._isDevTree(): # in a development tree
            docDir = join(self._getKomodoBitsDir(), "doc")
        else:
            if sys.platform == "win32":
                # mozBinDir: <installdir>/lib/mozilla
                # docDir:    <installdir>/doc
                docDir = join(dirname(dirname(self.get_mozBinDir())), "doc")
            elif sys.platform == "darwin":
                # mozBinDir: /Applications/Komodo.app/Contents/MacOS
                # docDir:    /Applications/Komodo.app/Contents/Resources/en.lproj/KomodoHelp
                docDir = join(dirname(self.get_mozBinDir()),
                              "Resources", "en.lproj", "KomodoHelp")
            else:
                # mozBinDir: <installdir>/lib/mozilla
                # docDir:    <installdir>/share/doc
                docDir = join(dirname(dirname(self.get_mozBinDir())),
                              "share", "doc")
        return docDir

    def get_installDir(self):
        # mozBinDir (Mac OS X): <installdir>/Contents/MacOS
        # mozBinDir (others):   <installdir>/lib/mozilla
        installDir = dirname(dirname(self.get_mozBinDir()))
        return installDir

    def get_binDir(self):
        if self._isDevTree(): # in a development tree
            binDir = join(self._getKomodoBitsDir(), "stub")
        else:
            if sys.platform == "win32":
                binDir = self.get_installDir()
            elif sys.platform == "darwin":
                binDir = self.get_mozBinDir()
            else:
                binDir = join(self.get_installDir(), "bin")
        return binDir

#XXX Coming later. See specs/tech/install_layout.txt
##    def get_etcDir(self):
##    def get_docDir(self):

    def get_binDBGPDir(self):
        return os.path.join(self.get_supportDir(), "dbgp", "bin")
    def get_perlDBGPDir(self):
        return os.path.join(self.get_supportDir(), "dbgp", "perllib")
    def get_pythonDBGPDir(self):
        return os.path.join(self.get_supportDir(), "dbgp", "pythonlib")


if __name__ == "__main__":
    koDirSvc = components.classes['@activestate.com/koDirs;1']\
               .getService(components.interfaces.koIDirs)
    dnames = ["userDataDir", "hostUserDataDir", "commonDataDir",
              "factoryCommonDataDir", "supportDir", "mozBinDir",
              "binDBGPDir", "perlDBGPDir", "pythonDBGPDir"]
    for dname in dnames:
        try:
            print "koIDirs.%s: %s" % (dname, getattr(koDirSvc, dname))
        except:
            print "koIDirs.%s: <error retrieving>" % dname

