#!python
# Copyright (c) 2000-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

# The implementation of the Komodo Run Command Service 

import os
import sys
import re
import tempfile
import threading
import process
import types
import urllib

from xpcom import components, nsError, ServerException, COMException, _xpcom
import logging
import which
from koTreeView import TreeView
import koprocessutils


#---- globals

_gLastErrorSvc = None
log = logging.getLogger("run")


#---- internal support stuff

class _NotifyOnTermination(threading.Thread):
    def __init__(self, process_, termListener, *args, **kwargs):
        threading.Thread.__init__(self, *args, **kwargs)
        self._termListener = termListener
        self._process = process_
    def run(self):
        try:
            retval = self._process.wait(self._process.INFINITE)
        except process.ProcessError, ex:
            if ex.errno == self._process.WAIT_TIMEOUT:
                # Inexplicably get this on WinNT.
                # See http://bugs.activestate.com/Komodo/show_bug.cgi?id=23176
                # Workaround: log the error but otherwise let it pass.
                log.error("Unexpected WAIT_TIMEOUT on .wait(INFINITE):"+str(ex))
                # If we return an _error_ code here, then the user gets
                # spurious warnings for some commands (e.g. PDK builds).
                # It is preferable to default to success.
                retval = 0
        self._termListener.onTerminate(retval)


class _KoRunProcessProxy(process.ProcessProxy):
    _com_interfaces_ = [components.interfaces.koIRunProcess]

    def wait(self, timeout):
        # Call the baseclass' wait() and set errors on koILastError as
        # appropriate.
        try:
            retval = process.ProcessProxy.wait(self, timeout)
        except process.ProcessError, ex:
            _gLastErrorSvc.setLastError(ex.errno, str(ex))
            raise
        return retval

    def readStdout(self):
        return self.stdout.read()
    def readStderr(self):
        return self.stderr.read()
            

#---- run service components

class KoInterpolationQuery:
    _com_interfaces_ = components.interfaces.koIInterpolationQuery
    _reg_desc_ = "Interpolation Query"
    _reg_clsid_ = "{6F169CEE-D4E1-477f-9AFA-69E82D139714}"
    _reg_contractid_ = "@activestate.com/koInterpolationQuery;1"

    def __init__(self, id, code, question, mruName=None, answer=None,
                 necessary=0, isPassword=False):
        self.id = id
        self.code = code
        self.question = question
        self.mruName = mruName
        self.answer = answer
        # A query is "necessary" if (1) it comes from an "ask" or
        # "askpass" code or (2) it comes from an "orask" modifier for
        # which no value could be automatically determined.
        self.necessary = necessary
        self.isPassword = isPassword

    def __repr__(self):
        answer_str = (self.answer and (" %s" % self.answer) or "")
        return "<KoInterpolationQuery %r: %s?%s>" \
               % (self.id, self.question, answer_str)


class KoInterpolationService:
    _com_interfaces_ = components.interfaces.koIInterpolationService
    _reg_desc_ = "Interpolation Service"
    _reg_clsid_ = "{C94C3130-A07E-4840-877B-E03D2F4BC872}"
    _reg_contractid_ = "@activestate.com/koInterpolationService;1"

    def __init__(self):
        self._prefSvc = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService)
        self._prefs = self._prefSvc.prefs
        self._userPath = koprocessutils.getUserEnv()["PATH"].split(os.pathsep)
        self._koDirSvc = components.classes["@activestate.com/koDirs;1"].\
            getService(components.interfaces.koIDirs)

        global _gLastErrorSvc
        if not _gLastErrorSvc:
            _gLastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                         .getService(components.interfaces.koILastErrorService)

    def _FindInterpreterOnPath(self, interp):
        """Return a full path to the first occurence of the given interpreter
        on the path.

        Return the empty string if none is found.
        """
        if sys.platform.startswith("win"):
            exts = ['.exe']
        else:
            exts = None
        if interp == "wish":
            # ActiveTcl does not put 'wish' in AppPaths but puts 'tclsh'
            # there.
            try:
                return which.which("wish", path=self._userPath)
            except which.WhichError:
                tclshs = which.whichall("tclsh", exts=exts,
                                        path=self._userPath)
                if sys.platform.startswith("win"):
                    wishBase = "wish.exe"
                else:
                    wishBase = "wish"
                for tclsh in tclshs:
                    wish = os.path.join(os.path.dirname(tclsh), wishBase)
                    if os.path.exists(wish):
                        return wish
                else:
                    return ""
        else:
            try:
                return which.which(interp, exts=exts, path=self._userPath)
            except which.WhichError:
                return ""

    def _GetInterpreter(self, interp, lang):
        """Return the full path to the given interpreter or raise a suitable
        ValueError.
        """
        code = "(%s)" % interp
        prefName = interp+"DefaultInterpreter"
        if self._prefs.hasStringPref(prefName):
            path = self._prefs.getStringPref(prefName)
        if not path:
            path = self._FindInterpreterOnPath(interp)
        
        if not path:
            raise ValueError("This command string includes '%%(%s)', but "\
                             "no '%s' interpreter could be found. You can "\
                             "configure this by visiting the %s "\
                             "preferences panel in Edit->Preferences."\
                             % (interp, interp, lang))
        else:
            return path

    def _getGuid(self):
        uuidGenerator = components.classes["@mozilla.org/uuid-generator;1"].getService(components.interfaces.nsIUUIDGenerator)
        guid = uuidGenerator.generateUUID()
        # must convert nsIDPtr to string first
        return str(guid)[1:-1] # strip off the {}'s

    def _getCodeMap(self, fileName=None, lineNum=None, word=None,
                    selection=None, projectFile=None):
        # Define the interpolation mapping.
        #   This is a mapping from interpolation code to the replacement
        #   value in the command. The "replacement" may be one of the
        #   following:
        #      - None; '%<code>' is left as is
        #      - a string; replaces '%<code>' in the command
        #      - a ValueError; exception string will be passed back through
        #        XPCOM and displayed as an error to the user
        #      - a callable object which either returns None or a string, or
        #        raises a ValueError
        codeMap = {
            '%': '%',
            'b': ValueError("The command string includes %b, but there is no current file"),
            'f': ValueError("The command string includes %f, but there is no current file"),
            'F': ValueError("The command string includes %F, but there is no current file"),
            'd': ValueError("The command string includes %d, but there is no current file"),
            'D': ValueError("The command string includes %D, but there is no current file"),
            'L': ValueError("The command string includes %L, but there is no current file"),
            'w': ValueError("The command string includes %w, but there is no selection or word under cursor"),
            'W': ValueError("The command string includes %W, but there is no selection or word under cursor"),
            's': ValueError("The command string includes %s, but there is no selection"),
            'S': ValueError("The command string includes %S, but there is no selection"),
            'P': ValueError("The command string includes %P, but there is no active project"),
            'p': ValueError("The command string includes %p, but there is no active project"),
            'perl':   lambda interp='perl',   lang='Perl':   self._GetInterpreter(interp, lang),
            'php':    lambda interp='php',    lang='PHP':    self._GetInterpreter(interp, lang),
            'python': lambda interp='python', lang='Python': self._GetInterpreter(interp, lang),



            'ruby':   lambda interp='ruby',   lang='Ruby':   self._GetInterpreter(interp, lang),
            'tclsh':  lambda interp='tclsh',  lang='Tcl':    self._GetInterpreter(interp, lang),
            'wish':   lambda interp='wish',   lang='Tcl':    self._GetInterpreter(interp, lang),
            'guid':   self._getGuid,

            # The code are handled specially in _doInterpolate1().
            'tabstop': None,
            'ask': None,
            'askpass': None,
            'date': None,
            'pref': None,
            'path': None,
            'debugger': None,
            }
        
        errorMsg = ""
        browser = self._prefs.getStringPref('browser')
        if not browser:
            # XXX this may not be the best way to fix getting a default
            # browser, but we have no method for that if the pref
            # is not set
            koWebbrowser = components.classes['@activestate.com/koWebbrowser;1'].\
                       getService(components.interfaces.koIWebbrowser)
            browsers = koWebbrowser.get_possible_browsers()
            if browsers:
                # Check that we have valid browser path
                if os.path.exists(os.path.abspath(browsers[0])):
                    browser = browsers[0]
                else:
                    errorMsg = "The command string includes %(browser), but the default browser is incorrectly set. Reset your default browser settings within your browser, or configure a default browser in Komodo's 'Web' Preference panel."
        if errorMsg or not browser:
            if not errorMsg:
                errorMsg = "The command string includes %(browser), but there is no default browser set. You can configure a default browser in Komodo's 'Web' Preference panel. Go to 'Edit | Preferences'."
            codeMap['browser'] = ValueError(errorMsg)
        else:
            if ' ' in browser and browser[0] != '"':
                browser = '"%s"' % browser
            codeMap['browser'] = browser

        from os.path import basename, dirname, splitext
        if fileName is not None:
            codeMap['b'] = splitext(basename(fileName))[0]
            codeMap['f'] = basename(fileName)
            codeMap['F'] = fileName
            codeMap['d'] = basename(dirname(fileName))
            codeMap['D'] = dirname(fileName)
        if lineNum > 0:
            codeMap['L'] = str(lineNum)
        if word:
            codeMap['w'] = word
            codeMap['W'] = word
        if projectFile:
            codeMap['P'] = projectFile
            codeMap['p'] = dirname(projectFile)
        if selection:
            codeMap['w'] = selection
            codeMap['s'] = selection
            # urlib quote does not support unicode strings
            if isinstance(selection, unicode):
                try:
                    selection = selection.encode()
                except UnicodeEncodeError:
                    selection = selection.encode('utf-8')
            codeMap['S'] = urllib.quote_plus(selection)
            codeMap['W'] = urllib.quote_plus(selection)
        return codeMap

    def _getCodeRes(self, codeMap, bracketed):
        """Return a list of regular expressions suitable for parsing
        code blocks out of strings.
        
        Want a regex that will match things like the following:
            %s  %a  %S  %(S) %(perl) %(python:orask) %perl
            %(python:orask:Which Python)
            %(w:else:Spam and Eggs)
            %(ask:Search for) %(ask:Search for:default value) %ask
        or, iff "bracketed":
            [[%s]] [[%(perl)]] [[%ask:Search for]]
        """
        allCharCodes = [re.escape(k) for k in codeMap.keys()]
        if bracketed:
            codeReStrs = [
                # Parse bracketed codes _with_ parentheses:
                #   [[%(w)]]  [[%(ask:Search for:default value)]]
                """\[\[\s*                          # opening bracketing
                    %%(                             # a leading percent char
                        \(
                            (?P<code>%s)            # the code
                            (?P<backref>\d+)?       # an optional back ref num
                            (:(?P<field1>[^:]*?)    # an optional field 1
                                (:(?P<field2>.*?))? # an optional field 2
                            )?
                        \)
                    )
                \s*\]\]""" % '|'.join(allCharCodes),
                # Parse bracketed codes _without_ parentheses:
                #   [[%w]]  [[%ask:Search for:default value]]
                """\[\[\s*                          # opening bracketing
                    %%(                             # a leading percent char
                        (?P<code>%s)                # the code
                        (?P<backref>\d+)?           # an optional back ref num
                        (:(?P<field1>[^:]*?)        # an optional field 1
                            (:(?P<field2>.*?))?     # an optional field 2
                        )?
                    )
                \]\]""" % '|'.join(allCharCodes),
            ]
        else:
            codeReStrs = [
                # Parse codes _with_ parentheses:
                #   %(w), %(ask:Search for:default value)
                """%%(                          # a leading percent char
                    \(
                        (?P<code>%s)            # the code
                        (?P<backref>\d+)?       # an optional back ref num
                        (:(?P<field1>[^:]*?)    # an optional field 1
                            (:(?P<field2>.*?))? # an optional field 2
                        )?
                    \)
                )""" % '|'.join(allCharCodes),
                # Parse codes _without_ parentheses:
                #   %w, %ask:Search:default
                """%%(                          # a leading percent char
                    (?P<code>%s)                # the code
                    (?P<backref>\d+)?           # an optional back ref num
                    (:(?P<field1>[^\s:]*)       # an opt field 1 (no spaces)
                        (:(?P<field2>[^\s]*))?  # an opt field 2 (no spaces)
                    )?
                )""" % '|'.join(allCharCodes),
            ]
        codeRes = [re.compile(s, re.VERBOSE) for s in codeReStrs]
        return codeRes

    def _doInterpolate1(self, s, codeMap, codeRes, prefSet, queries,
                        backrefs):
        """Do the first step interpolation on string "s". Return the result
        and add queries to "queries".

        "s" is the string to interpolate.
        "codeMap" is a mapping of the interpolation codes to their
            replacement.
        "codeRes" is a list of regular expressions for parsing out
            interpolation code blocks.
        "prefSet" is a koIPreferenceSet instance used for processing
            %(pref) interpolation codes. It can be None to indicate that
            the global preference set should be used.
        "queries" is a list of existing KoInterpolationQuery objects
            from interpolation codes in previous strings in the same set
            of strings to interpolate. New queries from "s" should be
            appended to this list.
        "backrefs" is a dictionary of interp code back references. This
            dict a mapping from the code name and backref number to the
            value for that code. The value is either a replacement
            string or a KoInterpolationQuery id.
            {"w1": "foo",
             "guid1": "8BE89689-2A82-433b-9D88-726737ED2547",
             "ask1": "%(__query1__)"}
        
        A note on "queries": The list of queries is ultimately returned
        to the JavaScript code using this service to present a query dialog
        to the user. Queries can result from either the %(ask) code or
        the ":orask" modifier on any of the other codes. Normally the latter,
        ":orask", case will only result in a query if no value could be
        determined for the main code, e.g. %(F:orask) with no current file.
        However, we would like these orask's to _also_ generate a query even
        if a value was determined for them _iff_ some other code requires
        a query.
        """
        #print "_doInterpolate1(s=%r, codeMap, codeRes, prefSet, "\
        #      "queries=%r, backrefs=%r)" % (s, queries, backrefs)
        idx = 0
        i1s = ""  # the interpolated result
        while s and idx < len(s):
            # Find the _first_ match with the available regexes.
            # Otherwise "%(f)" is found first in "echo %f %(f)" and the
            # first "%f" is never matched.
            match = None
            for codeRe in codeRes:
                m = codeRe.search(s, idx)
                if m and (match is None or m.start() < match.start()):
                    match = m
            if match is None:
                #print "_doInterpolate1: %r did not match" % s[idx:]
                i1s += s[idx:] # tack on the remainder
                break
            #print "_doInterpolate1: %r matched: %r groups=%r"\
            #      % (s[idx:], s[match.start():match.end()], match.groupdict())

            i1s += s[idx:match.start()]

            # The meaning of "field1" and "field2" depends on the "code".
            hit = match.groupdict()
            code = hit["code"]
            defaultValue = ""
            defaultAnswer = ""
            question = ""
            if hit["backref"]:
                backref = code+hit["backref"]
            else:
                backref = None
            if code in ("ask", "askpass"):
                question = hit["field1"]
                defaultAnswer = hit["field2"]
            elif code == "date":
                format = hit["field1"]
                if hit["field2"] is not None:
                    # Allow %(date:%H:%M:%S) to yield a format="%H:%M:%S".
                    format += ":"+hit["field2"]
            elif code =="pref":
                prefName = hit["field1"]
                defaultValue = hit["field2"]
            elif code =="path":
                pathValue = hit["field1"]
            elif code =="debugger":
                name = hit["field1"]
            elif code =="tabstop":
                defaultValue = hit["field1"]
            else:
                modifier = hit["field1"]
                if modifier == "orask":
                    question = hit["field2"]
                elif modifier == "else":
                    default = hit["field2"] or ""
                elif modifier is not None:
                    errmsg = "Unexpected modifier '%s' in '%s'"\
                             % (modifier, s[match.start():match.end()])
                    _gLastErrorSvc.setLastError(nsError.NS_ERROR_INVALID_ARG,
                                                errmsg)
                    raise ServerException(nsError.NS_ERROR_INVALID_ARG, errmsg)


            if backref and backref in backrefs:
                # This is a back-reference. Use the value of the
                # original reference.
                i1s += backrefs[backref]

            elif code in ("ask", "askpass"):
                # The "ask" code means that we will ask the user always.
                id = "%%(__query%d__)" % len(queries)
                if question:
                    mruName = str(question)
                else:
                    question = "Value %d" % len(queries)
                    mruName = "default"
                if code == "askpass":
                    mruName = None # never store passwords in an MRU

                # Interpolate the default answer. Only limitation is that
                # queries in the default answer are not allowed (i.e. no
                # %ask, %askpass or, "orask" modifier).
                if defaultAnswer and len(defaultAnswer) >= 2 \
                   and "%" in defaultAnswer:
                    answerCodeRes = self._getCodeRes(codeMap, False)
                    answerQueries = []
                    answerBackrefs = {}
                    iDefaultAnswer = self._doInterpolate1(
                        defaultAnswer, codeMap, answerCodeRes, prefSet,
                        answerQueries, answerBackrefs)
                    if not answerQueries: # Queries in query default not allowed.
                        defaultAnswer = iDefaultAnswer
                    
                    
                q = KoInterpolationQuery(id=id, code=code, question=question,
                                         mruName=mruName,
                                         answer=defaultAnswer,
                                         necessary=1,
                                         isPassword=(code == "askpass"))
                queries.append( q )
                i1s += id
                if backref:
                    backrefs[backref] = id
                
            elif code == "pref":
                if prefSet is None:
                    prefSet = self._prefSvc.effectivePrefs
                if prefSet.hasPref(prefName):
                    prefType = prefSet.getPrefType(prefName)
                    if prefType == "string":
                        value = prefSet.getStringPref(prefName)
                    elif prefType == "long":
                        value = '%d' % prefSet.getLongPref(prefName)
                    elif prefType == "double":
                        value = '%f' % prefSet.getDoublePref(prefName)
                    elif prefType == "boolean":
                        value = '%d' % prefSet.getBooleanPref(prefName)
                    else:
                        errmsg = "Unable to handle pref type '%s' for '%s'"\
                                 % (prefType, prefName)
                        _gLastErrorSvc.setLastError(nsError.NS_ERROR_INVALID_ARG,
                                                    errmsg)
                        raise ServerException(nsError.NS_ERROR_INVALID_ARG, errmsg)
                elif defaultValue:
                    value = defaultValue
                else:
                    errmsg = "Unable to get value for pref '%s'"\
                             % (prefName)
                    _gLastErrorSvc.setLastError(nsError.NS_ERROR_INVALID_ARG,
                                                errmsg)
                    raise ServerException(nsError.NS_ERROR_INVALID_ARG, errmsg)
                
                i1s += value
                if backref:
                    backrefs[backref] = value

            elif code == "path":
                if pathValue in ("userDataDir",
                                 "hostUserDataDir",
                                 "commonDataDir",
                                 "supportDir"):
                    value = getattr(self._koDirSvc, pathValue)
                else:
                    errmsg = "Unable to get path for '%s'"\
                             % (pathValue)
                    _gLastErrorSvc.setLastError(nsError.NS_ERROR_INVALID_ARG,
                                                errmsg)
                    raise ServerException(nsError.NS_ERROR_INVALID_ARG, errmsg)

                i1s += value
                if backref:
                    backrefs[backref] = value
            elif code == "tabstop":
                value = u"\xab%s\xbb" % (defaultValue or "")
                i1s += value
                if backref:
                    backrefs[backref] = value































            elif code == "date":
                import time
                if format is None:
                    format = self._prefs.getStringPref("defaultDateFormat")
                t = time.localtime(time.time())
                value = time.strftime(format, t)
                i1s += value
                if backref:
                    backrefs[backref] = value

            elif modifier == "orask":
                # The "orask" modifier means we only ask if no value in
                # the given context could be determined. We generate a
                # "query" in either case, but the query is only "necessary"
                # if no value could be determined.
                id = "%%(__query%d__)" % len(queries)
                if question:
                    mruName = str(question)
                else:
                    question = "Value %d" % len(queries)
                    mruName = "default"

                answer = codeMap[code]
                if callable(answer):
                    answer = answer()
                elif isinstance(answer, Exception):
                    answer = None
                necessary = answer is None

                q = KoInterpolationQuery(id=id, code=code, question=question,
                                         mruName=mruName, answer=answer,
                                         necessary=necessary,
                                         isPassword=False)
                queries.append( q )
                i1s += id
                if backref:
                    backrefs[backref] = id

            else:
                replacement = codeMap[code]
                if replacement is None:
                    if modifier == "else":
                        replacement = default
                    else:
                        replacement = s[match.start():match.end()]
                if callable(replacement):
                    replacement = replacement()
                elif isinstance(replacement, Exception):
                    if modifier == "else":
                        replacement = default
                    else:
                        raise replacement
                i1s += replacement
                if backref:
                    backrefs[backref] = replacement
            idx = match.end()
        return i1s

    def Interpolate1(self, strings, bracketedStrings, fileName, lineNum, word, selection,
                     projectFile, prefSet):
        try:
            #print "Interpolate1(strings=%r, fileName=%r, lineNum=%r, word=%r, "\
            #      "selection=%r, projectFile, prefSet, bracketed=%r)"\
            #      % (strings, fileName, lineNum, word, selection, bracketed)
            codeMap = self._getCodeMap(fileName, lineNum, word, selection, projectFile)
            codeRes = None
            bracketedCodeRes = None
            if strings:
                codeRes = self._getCodeRes(codeMap, 0)
            if bracketedStrings:
                bracketedCodeRes = self._getCodeRes(codeMap, 1)
    
            # Walk through the given strings replacing codes with their
            # values. Code blocks with one of the "ask" modifiers are tagged
            # and returned in the list of returned KoInterpolationQuery's.
            queries = []
            backrefs = {}
            try:
                i1strings = []
                for s in strings:
                    i1s = self._doInterpolate1(s, codeMap, codeRes, prefSet,
                                               queries, backrefs)
                    i1strings.append(i1s)
                    i1strings.append(i1s) # see NOTE in koIRunService on doubling strings
                for s in bracketedStrings:
                    i1s = self._doInterpolate1(s, codeMap, bracketedCodeRes, prefSet,
                                               queries, backrefs)
                    i1strings.append(i1s)
                    i1strings.append(i1s) # see NOTE in koIRunService on doubling strings
            except ValueError, ex:
                log.exception(ex)
                _gLastErrorSvc.setLastError(nsError.NS_ERROR_INVALID_ARG,
                                            ex.args[0])
                raise ServerException(nsError.NS_ERROR_INVALID_ARG,
                                      ex.args[0])
    
            # If any queries are necessary then we pass them all back. If none
            # of them are necessary then we can interpolate them all and pass
            # none of them back.
            necessaryQueries = [q for q in queries if q.necessary]
            if necessaryQueries:
                return queries, i1strings
            else:
                i2strings = self.Interpolate2(i1strings, queries)
                return [], i2strings
        except Exception, e:
            log.exception(e)
            raise

    def Interpolate2(self, i1strings, queries):
        #print "Interpolate2(i1strings=%r, queries=%r)"\
        #      % (i1strings, queries)
        # Precondition: all queries have an answer.
        try:
            for q in queries: assert q.answer is not None
    
            i2strings = []
            for i1string in i1strings[0:len(i1strings):2]: # skip duplicates
                forUse = forDisplay = i1string
                for q in queries:
                    answer = q.answer
                    if q.code in ('S', 'W'):
                        # These are codes that are supposed to URL-escape selected
                        # values. Extend that same functionality to such values
                        # entered in the Interpolation Query dialog.
                        # urlib quote does not support unicode strings
                        if isinstance(answer, unicode):
                            try:
                                answer = answer.encode()
                            except UnicodeEncodeError:
                                answer = answer.encode('utf-8')
                        answer = urllib.quote_plus(answer)
                    forUse = forUse.replace(q.id, answer)
                    if q.isPassword:
                        forDisplay = forDisplay.replace(q.id, '*'*len(answer))
                    else:
                        forDisplay = forDisplay.replace(q.id, answer)
                i2strings += [forUse, forDisplay]
    
            return i2strings
        except Exception, e:
            log.exception(e)
            raise

    def GetBlockOffsets(self, s, bracketed):
        codeMap = self._getCodeMap()
        codeRes = self._getCodeRes(codeMap, bracketed)

        offsets = []
        idx = 0
        while idx < len(s):
            for codeRe in codeRes:
                match = codeRe.search(s, idx)
                if match:
                    offsets += [match.start(), match.end()]
                    break
            else:
                break # no more code blocks
            idx = match.end()

        #print "GetBlockOffsets(s=%r): %r" % (s, offsets)
        return offsets
        


class KoRunService:
    _com_interfaces_ = components.interfaces.koIRunService
    _reg_desc_ = "Run Service"
    _reg_clsid_ = "{8BE89689-2A82-433b-9D88-726737ED2547}"
    _reg_contractid_ = "@activestate.com/koRunService;1"

    def __init__(self):
        global _gLastErrorSvc
        _gLastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                         .getService(components.interfaces.koILastErrorService)

        # Setup for sending status bar messages in sub-threads.
        self._observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self._observerSvcProxy = _xpcom.getProxyForObject(None,
            components.interfaces.nsIObserverService, self._observerSvc,
            _xpcom.PROXY_ALWAYS | _xpcom.PROXY_SYNC)
        self._sm = components.classes["@activestate.com/koStatusMessage;1"]\
            .createInstance(components.interfaces.koIStatusMessage)

        if sys.platform.startswith("win"):
            if os.environ.has_key("SHELL"):
                shell = os.environ["SHELL"]
            elif os.environ.has_key("ComSpec"):
                shell = os.environ["ComSpec"]
            else:
                #XXX Would be nice to get full path here so can use
                #    os.spawn* calls.
                shell = "command.com"
            shellArgs = ["/C"]
        elif sys.platform == "darwin":
            shell = "/bin/sh"
            shellArgs = []
        else:
            shell = "xterm"
            shellArgs = ["-e", "/bin/sh"]
        self.shellArgs = [shell] + shellArgs
            
    
    def Encode(self, command, cwd, env, insertOutput,
               operateOnSelection, doNotOpenOutputWindow, runIn,
               parseOutput, parseRegex, showParsedOutputList):
        """Encode the command (with all options) into a string.

        This will be displayed in the "Recent Commands" MRU so important
        information should be first. Plan:
            <command> <dict with options>
                - escape '{' and '}' in <command>
        """
        encoded = command.replace('{', '\\{')\
                         .replace('}', '\\}')
        # Add an option to the optdict if its value is other than the
        # default.
        optdict = {}
        if cwd:
            optdict['cwd'] = cwd
        if env:
            optdict['env'] = env
        if insertOutput:
            optdict['insertOutput'] = insertOutput
        if operateOnSelection:
            optdict['operateOnSelection'] = operateOnSelection
        if doNotOpenOutputWindow:
            optdict['doNotOpenOutputWindow'] = doNotOpenOutputWindow
        if runIn != 'command-output-window':
            optdict['runIn'] = runIn
        if parseOutput:
            optdict['parseOutput'] = parseOutput
        if parseRegex:
            optdict['parseRegex'] = parseRegex
        if showParsedOutputList:
            optdict['showParsedOutputList'] = showParsedOutputList
        if optdict:
            encoded += ' ' + repr(optdict)
        return encoded

    def Decode(self, encoded):
        """Decode an encoded command (see .Encode()).
        
        Want to support old encoding style (from Komodo 1.x) for
        backward compatibility.  From the old docs the form is:
           [<options letters] <command>
        Options letters are (these are the same as thier accesskeys in
        run.xul):
           i == insertOutput 
           l == operateOnSelection
        """
        optdict = { # defaults
            'cwd': '',
            'env': '',
            'insertOutput': 0,
            'operateOnSelection': 0,
            'doNotOpenOutputWindow': 0,
            'runIn': 'command-output-window',
            'parseOutput': 0,
            'parseRegex': '',
            'showParsedOutputList': 0,
        }

        # First see if this is possibly an old-style encoding.
        oldStyleRe = re.compile(
            "^\[(?P<insertOutput>i)?(?P<operateOnSelection>l)?\] "\
            "(?P<command>.*)$")
        match = oldStyleRe.match(encoded)
        if match:
            if match.group('insertOutput') is not None:
                optdict['insertOutput'] = 1
            if match.group('operateOnSelection') is not None:
                optdict['operateOnSelection'] = 1
            command = match.group('command')

        # New-style encoding
        else:
            parts = encoded.split(' {', 1)
            if len(parts) == 1:
                command = parts[0]
                optdictstr = '{}'
            else:
                command, optdictstr = parts
                optdictstr = '{' + optdictstr
            command = command.replace('\\{', '{')\
                             .replace('\\}', '}')
            try:
                optupdate = eval(optdictstr)
            except SyntaxError, ex:
                # This likely means that we are trying to decode an old
                # style encoded command as if it were a new-style
                # encoded command and the encoded command includes '{'
                # and/or '}' chars.
                command = encoded
            else:
                optdict.update(optupdate)

        return (command, optdict['cwd'], optdict['env'],
                optdict['insertOutput'], optdict['operateOnSelection'],
                optdict['doNotOpenOutputWindow'], optdict['runIn'],
                optdict['parseOutput'], optdict['parseRegex'],
                optdict['showParsedOutputList'])

    def _getEnvDictFromString(self, env):
        """Convert this:
            "VAR1=value1\nVAR2=value2\n..."
        to this:
            {"VAR1": "value1", "VAR2": "value2", ...}
        """
        envDict = {}
        for piece in env.split('\n'):
            if not piece: continue
            equalSign = piece.find('=')
            if equalSign == -1:
                log.warn("no '=' in env string line (skipping): %r" % piece)
                continue
            envDict[piece[:equalSign]] = piece[equalSign+1:]
        return envDict

    def _mergeEnvWithParent(self, envDict):
        """Merge the given environment variables with parent process' environ.

        Note: koIUserEnviron is used to avoid picking up the current
              Komodo process' environment changes from getting in the way.
        """
        userEnvSvc = components.classes["@activestate.com/koUserEnviron;1"].getService()
        userEnvDict = {}
        for piece in userEnvSvc.GetEnvironmentStrings():
            equalSign = piece.find('=')
            userEnvDict[piece[:equalSign]] = piece[equalSign+1:]
        userEnvDict.update(envDict)
        return userEnvDict

    def RunInTerminal(self, command, cwd, env, terminal, termListener,
                      input=None):
        """Run the given command connected to the given terminal.

        Optionally an "input" string can be provided. If it is then the
        terminal's stdin is *not* used. Instead the given input string
        is passed to the child and that pipe is closed.

        A termination listener, "termListener", can ge registered to
        receive notification when the child terminates and with what
        exit value.
        """
        #print "KoRunService.RunInTerminal(command=%r, cwd=%r, env=%r, "\
        #      "terminal=%r, termListener=%r, input=%r)"\
        #      % (command, cwd, env, terminal, termListener, input)
        if not cwd: # Allow cwd="".
            cwd = None
        if cwd and cwd.startswith('~'):
            cwd = os.path.expanduser(cwd)
        if not env:
            env = ""
        envDict = self._getEnvDictFromString(env)
        envDict = self._mergeEnvWithParent(envDict)
        # The process library requires that all env strings be ASCII
        # or all Unicode. We'll just make them all Unicode for now.
        #XXX This unicode conversion may not be necessary since
        #    _SaferCreateProcess in process.py.
        uEnvDict = {}
        for key, val in envDict.items():
            #XXX Linux may have trouble with all unicode env
            #    strings. Perhaps have all ASCII on Linux.
            uEnvDict[unicode(key)] = unicode(val)
        envDict = uEnvDict

        if input:
            try:
                p = _KoRunProcessProxy(command, cwd=cwd, env=envDict,
                                       stdout=terminal.stdout,
                                       stderr=terminal.stderr)
            except process.ProcessError, ex:
                _gLastErrorSvc.setLastError(ex.errno, str(ex))
                raise ServerException(nsError.NS_ERROR_FAILURE, str(ex))
            p.stdin.write(input)
            p.stdin.close()
        else:
            try:
                p = _KoRunProcessProxy(command, cwd=cwd, env=envDict,
                                       mode='b',  # XXX need to expose this
                                       stdin=terminal.stdin,
                                       stdout=terminal.stdout,
                                       stderr=terminal.stderr)
            except process.ProcessError, ex:
                _gLastErrorSvc.setLastError(ex.errno, str(ex))
                raise ServerException(nsError.NS_ERROR_FAILURE, str(ex))

        # If the user provides a termination listener (interface
        # koIRunTerminationListener), then setup a thread to notify of
        # termination of the child.
        if termListener:
            _termListenerProxy = _xpcom.getProxyForObject(None,
                components.interfaces.koIRunTerminationListener, termListener,
                _xpcom.PROXY_ALWAYS | _xpcom.PROXY_SYNC)
            t = _NotifyOnTermination(p, _termListenerProxy)
            t.start()
        # XXX What if no termination listener is provided? Does the loss
        #     of 'p' reference cause premature deletion of the Process?
        #     This might be moot when (if?) we start passing back a
        #     koIRunProcess object holding the Process object. Nope, it
        #     would not be moot if the user did not assign the returned
        #     process to a variable. Perhaps a _NotifyOnTermination()
        #     should *always* be created, holding that reference. This
        #     then *requires* that all child processes must close before
        #     Komodo, which sucks for, say, TkBuilder (*does* this imply
        #     that?).
        return p

    def _WaitAndCleanUp(self, child, command, scriptFileName=None,
                        inputFileName=None):
        retval = child.wait()

        if scriptFileName:
            try:
                os.unlink(scriptFileName)
            except OSError, ex:
                log.warn("Could not remove temporary script file '%s': %s"\
                           % (scriptFileName, ex))
        if inputFileName:
            try:
                os.unlink(inputFileName)
            except OSError, ex:
                log.warn("Could not remove temporary input file '%s': %s"\
                           % (inputFileName, ex))
        
        # Put appropriate message on the status bar.
        self._sm.category = "run_command"
        self._sm.msg = "'%s' returned %s." % (command, retval)
        self._sm.timeout = 3000
        if retval == 0:
            self._sm.highlight = 0
        else:
            self._sm.highlight = 1
        try:
            self._observerSvcProxy.notifyObservers(self._sm, 'status_message', None)
        except COMException, e:
            # do nothing: Notify sometimes raises an exception if (???)
            # receivers are not registered?
            pass

    def Run(self, command, cwd, env, console, input=None):
        """Run the command string in a new (console) window

        "console" is a boolean indicating whether the child should be
            spawned with a new console window or not.
        
        Optionally an "input" string can be provided. If it is then the
        command is run via "command < input", where the input is placed
        in a temporary file.
        """
        #print "KoRunService.Run(command=%r, cwd=%r, env=%r, "\
        #      "console=%r, input=%r)"\
        #      % (command, cwd, env, console, input)
        if not cwd: # Allow cwd="".
            cwd = None
        if cwd and cwd.startswith('~'):
            cwd = os.path.expanduser(cwd)
        if not env:
            env = ""
        envDict = self._getEnvDictFromString(env)
        envDict = self._mergeEnvWithParent(envDict)
        # The process library requires that all env strings be ASCII
        # or all Unicode. We'll just make them all Unicode for now.
        #XXX This unicode conversion may not be necessary since
        #    _SaferCreateProcess in process.py.
        uEnvDict = {}
        for key, val in envDict.items():
            uEnvDict[unicode(key)] = unicode(val)
        envDict = uEnvDict

        # Keep a pure 'command' for reporting status to the user.
        actualCommand = command

        if input:
            # Put the input into a temp file.
            inputFileName = tempfile.mktemp()
            inputFile = open(inputFileName, "w")
            inputFile.write(input)
            inputFile.close()
            if ' ' in inputFileName:
                actualCommand += ' < "%s"' % inputFileName
            else:
                actualCommand += ' < %s' % inputFileName
        else:
            inputFileName = None

        if console:
            # A shell script that will be run is created.  The script
            # adds the "Press any key to continue..." feature and
            # ensures that the correct return value is returned.
            if sys.platform.startswith("win"):
                scriptFileName = tempfile.mktemp() + ".bat"
                bat = open(scriptFileName, "w")
                bat.write("@echo off\n")
                # Escape '%' once for being in a batch file plus once for
                # being on a "call ..." line.
                actualCommand = actualCommand.replace("%", "%%%%")
                bat.write("call %s\n" % actualCommand)
                bat.write("set KOMODO_COMMAND_ERRORLEVEL=%ERRORLEVEL%\n")
                bat.write("pause\n")
                # Must "exit" with the last return value to ensure that
                # cmd.exe will return that value
                #XXX does the correct retval actually get returned?
                bat.write("exit %KOMODO_COMMAND_ERRORLEVEL%\n")
                bat.close()
                #XXX does this work on Win9x???
            elif sys.platform == "darwin":
                # We can use applescript to run a shell command in a new
                # Terminal.app window. The problem is that the relevant
                # AppleScript command does not wait for the command to
                # terminate. We need to resort to locks and temp files to
                # (a) wait for termination and (b) get the retval.
                #
                # 1. lockfile foo.lock
                # 2. run foo.applescript to run the following in the terminal:
                #    - run the user's command
                #    - write the retval to a temp file
                #    - rm -f foo.lock
                # 3. lockfile -1 foo.lock  # i.e. wait for foo.lock to get rm'd
                # 4. exit with the retval
                #
                def escapecmd(cmd):
                    return cmd.replace('"', '\\"').replace('$', '\\$')
                base = tempfile.mktemp()
                shEnvFile = base+".environ.sh"
                cshEnvFile = base+".environ.csh"
                data = {
                    "lockfile": base+".lock",
                    "retvalfile": base+".retval",
                    "applescriptfile": base+".applescript",
                    "cmd": escapecmd(actualCommand),
                    "cwdcmd": "",
                    "shenvfile": shEnvFile,
                    "cshenvfile": cshEnvFile,
                    "envcmd": (r"echo \$shell | grep csh >/dev/null "
                               r" && source %s "
                               r" || source %s" % (cshEnvFile, shEnvFile)),
                }
                if cwd:
                    data["cwdcmd"] = r'cd \"%s\";' % cwd
    
                def quotevalue(value):
                    if "'" in value:
                        for ch in "\\$!": # special shell chars to escape
                            value = value.replace(ch, '\\'+ch)
                        return '"%s"' % value
                    else:
                        return "'%s'" % value
                fenv = open(shEnvFile, "w")
                for name, value in envDict.items():
                    fenv.write('    %s=%s; export %s\n' % (name, quotevalue(value), name))
                fenv.close()
                fenv = open(cshEnvFile, "w")
                for name, value in envDict.items():
                    fenv.write('    setenv %s %s\n' % (name, quotevalue(value)))
                fenv.close()

                scriptFileName = base+".sh"
                script = open(scriptFileName, "w")
                script.write(r"""
LOCKFILE='%(lockfile)s'
RETVALFILE='%(retvalfile)s'
APPLESCRIPTFILE='%(applescriptfile)s'
SHENVFILE='%(shenvfile)s'
CSHENVFILE='%(cshenvfile)s'

rm -f $LOCKFILE
lockfile $LOCKFILE
cat >$APPLESCRIPTFILE <<HERE
tell app "Terminal"
    do script "clear; %(cwdcmd)s (%(envcmd)s; %(cmd)s); echo \$?>$RETVALFILE; echo Press RETURN to continue ...; /bin/sh -c read RETURN; rm -f $LOCKFILE; exit"
    delay 0.2
    activate
end tell
HERE
#echo debug: running AppleScript in $APPLESCRIPTFILE:
#cat $APPLESCRIPTFILE
#echo debug: which will source one of $SHENVFILE:
#cat $SHENVFILE
#echo debug: or $CSHENVFILE:
#cat $CSHENVFILE
osascript $APPLESCRIPTFILE
#echo debug: waiting for Terminal to finish...
lockfile -1 $LOCKFILE
RETVAL=`cat $RETVALFILE`
#echo debug: retval was \"$RETVAL\"
rm -f $RETVALFILE $LOCKFILE $SHENVFILE $CSHENVFILE $APPLESCRIPTFILE
#XXX Need Komodo to register with OSA (bug 37377).
#osascript -e 'tell app "Komodo" to activate'
exit $RETVAL""" % data)
                script.close()

            else:
                # Create a bash script that will be run.
                scriptFileName = tempfile.mktemp() + ".sh"
                script = open(scriptFileName, "w")
                script.write("#!/bin/sh\n")
                script.write("%s\n" % actualCommand)
                script.write("set KOMODO_COMMAND_RETVAL=$?\n")
                script.write("export KOMODO_COMMAND_RETVAL\n")
                if sys.platform.startswith("sunos"):
                    # Solaris /bin/sh is doesn't understand -n
                    script.write("echo Press ENTER to continue . . .\n")
                else:
                    script.write("echo -n Press ENTER to continue . . .\n")
                script.write("read KOMODO_DUMMY\n")
                script.write("unset KOMODO_DUMMY\n")
                script.write("exit $KOMODO_COMMAND_RETVAL\n")
                script.close()
            actualCommand = self.shellArgs + [scriptFileName]
            
            #print "RUN: actualCommand is '%s'" % actualCommand
            #print "RUN: -------- %s is as follows -------" % scriptFileName
            #fin = open(scriptFileName, 'r')
            #for line in fin.readlines():
            #    print line,
            #fin.close()
            #print "RUN: ---------------------------------"
            flags = process.Process.CREATE_NEW_CONSOLE
        else:
            scriptFileName = None
            flags = 0

        try:
            child = process.Process(actualCommand, cwd=cwd, env=envDict,
                                    flags=flags)
        except process.ProcessError, ex:
            _gLastErrorSvc.setLastError(ex.errno, str(ex))
            raise ServerException(nsError.NS_ERROR_FAILURE, str(ex))

        # The return value is passed to the status bar when the child
        # terminates. A separate thread is created to handle that so
        # this call can return immediately.
        t = threading.Thread(target=self._WaitAndCleanUp,
                             kwargs={'child': child,
                                     'command': command,
                                     'scriptFileName': scriptFileName,
                                     'inputFileName': inputFileName})
        t.start()

    def RunAndCaptureOutput(self, command, cwd, env, input=None):
        """Run the command and return the output.

        Optionally an "input" string can be provided. If it is then
        the given input string is passed to the child and that pipe is
        closed.
        """
        #print "KoRunService.RunAndCaptureOutput(command=%r, cwd=%r, "\
        #      "env=%r, input=%r)" % (command, cwd, env, input)
        if not cwd: # Allow cwd="".
            cwd = None
        if cwd and cwd.startswith('~'):
            cwd = os.path.expanduser(cwd)
        if not env:
            env = ""
        envDict = self._getEnvDictFromString(env)
        envDict = self._mergeEnvWithParent(envDict)
        # The process library requires that all env strings be ASCII
        # or all Unicode. We'll just make them all Unicode for now.
        #XXX This unicode conversion may not be necessary since
        #    _SaferCreateProcess in process.py.
        uEnvDict = {}
        for key, val in envDict.items():
            uEnvDict[unicode(key)] = unicode(val)
        envDict = uEnvDict

        try:
            child = process.ProcessProxy(command, cwd=cwd, env=envDict)
        except process.ProcessError, ex:
            _gLastErrorSvc.setLastError(ex.errno, str(ex))
            raise ServerException(nsError.NS_ERROR_FAILURE, str(ex))

        if input:
            child.stdin.write(input)
            child.stdin.close()
        retval = child.wait()
        output = child.stdout.read()
        error = child.stderr.read()
        child.close()

        return retval, output, error

    def _WaitAndNotify(self, child, command):
        retval = child.wait(child.INFINITE)

        # Notify any listener of child termination.
        try:
            #XXX Dunno if need to pass 'child' in the notification here.
            self._observerSvcProxy.notifyObservers(child, 'run_terminated', command)
        except COMException, e:
            # Do nothing: Notify sometimes raises an exception if (???)
            # no receiver are registered?
            pass

    def RunAndNotify(self, command, cwd, env, input=None):
        """Run the given command, returning the process object
        immediately, and notify (via nsIObserver) when the process
        terminates.
        """
        #print "KoRunService.RunAndNotify(command=%r, cwd=%r, "\
        #      "env=%r, input=%r)" % (command, cwd, env, input)
        if not cwd: # Allow cwd="".
            cwd = None
        if cwd and cwd.startswith('~'):
            cwd = os.path.expanduser(cwd)
        if not env:
            env = ""
        envDict = self._getEnvDictFromString(env)
        envDict = self._mergeEnvWithParent(envDict)
        # The process library requires that all env strings be ASCII
        # or all Unicode. We'll just make them all Unicode for now.
        #XXX This unicode conversion may not be necessary since
        #    _SaferCreateProcess in process.py.
        uEnvDict = {}
        for key, val in envDict.items():
            uEnvDict[unicode(key)] = unicode(val)
        envDict = uEnvDict

        try:
            child = _KoRunProcessProxy(command, cwd=cwd, env=envDict)
        except process.ProcessError, ex:
            _gLastErrorSvc.setLastError(ex.errno, str(ex))
            raise ServerException(nsError.NS_ERROR_FAILURE, str(ex))

        if input:
            child.stdin.write(input)
            child.stdin.close()

        t = threading 
        t = threading.Thread(target=self._WaitAndNotify,
                             kwargs={'child': child, 'command': command})
        t.start()
        return child


class KoRunEnvView(TreeView):
    _com_interfaces_ = [components.interfaces.koIRunEnvView,
                        components.interfaces.nsITreeView]
    _reg_clsid_ = "{0E3559E7-093D-47d1-ABC6-717FADB0D6C8}"
    _reg_contractid_ = "@activestate.com/koRunEnvView;1"
    _reg_desc_ = "Komodo Run Command Environment Variables Tree View"

    def __init__(self):
        TreeView.__init__(self, debug=0)
        self._data = {}
        # A list of the current variable names (i.e. self._data.keys())
        # giving the order in the table.
        self._order = []
        # A string key indicating by which column the env table is
        # sorted.
        self._sortedBy = None

    def get_rowCount(self):
        return len(self._data.keys())

    def getCellText(self, row, column):
        """Get the value for the specified cell.
        "row" is an index into the current order (self._order).
        "col" is an 'id' of the <treecol> element. Here
            "env-variable" or "env-value".
        """

        col = column.id



        try:
            variable = self._order[row]
        except IndexError:
            log.error("no %sth row of data" % row)
            return ""

        cell = None
        if col == "env-variable":
            cell = variable
        elif col == "env-value":
            cell = self._data[variable]
        else:
            log.error("unknown run environment variables column id: '%s'" % col)
            return ""

        if type(cell) not in (types.StringType, types.UnicodeType):
            cell = str(cell)
        return cell

    def Set(self, variable, value):
        if sys.platform.startswith("win"):
            variable = variable.upper()
        self._tree.beginUpdateBatch()
        if not self._data.has_key(variable):
            self._order.append(variable)
            self._sortedBy = None
            self._tree.rowCountChanged(len(self._data)-2, 1)
        self._data[variable] = value
        self._tree.invalidate()  #XXX invalidating too much here?
        self._tree.endUpdateBatch()

    def Delete(self, index):
        del self._data[self._order[index]]
        del self._order[index]
        self._tree.beginUpdateBatch()
        self._tree.invalidate()  #XXX invalidating too much here?
        self._tree.endUpdateBatch()

    def Have(self, variable):
        """Return true iff have a datum with the given variable name."""
        if sys.platform.startswith("win"):
            variable = variable.upper()
        return self._data.has_key(variable)

    def Index(self, variable):
        if sys.platform.startswith("win"):
            variable = variable.upper()
        try:
            return self._order.index(variable)
        except ValueError:
            return -1

    def GetVariable(self, index):
        return self._order[index]
    def GetValue(self, index):
        return self._data[self._order[index]]

    def Sort(self, sortBy):
        """Sort the current data by the given key. If already sorted by this
        key then reverse the sorting order."""
        if self._sortedBy == sortBy:
            self._order.reverse()
        else:
            if sortBy == 'env-variable':
                self._order.sort()
            elif sortBy == 'env-value':
                self._order.sort(lambda a,b,data=self._data,sortBy=sortBy:
                                    cmp(data[a], data[b])
                                )
            else:
                log.error("Cannot sort results by: '%s'" % sortBy)
                return
        self._sortedBy = sortBy
        self._tree.beginUpdateBatch()
        self._tree.invalidate()
        self._tree.endUpdateBatch()

    def GetEnvironmentStrings(self):
        vars = ["%s=%s" % item for item in self._data.items()]
        return "\n".join(vars)

    def SetEnvironmentStrings(self, s):
        self._data = {}
        for envvar in s.split('\n'):
            if envvar: # skip empty lines
                var, val = envvar.split('=', 1)
                self._data[var] = val
        self._order = self._data.keys()
        self._sortedBy = None
        if self._tree:
            self._tree.beginUpdateBatch()
            self._tree.invalidate()
            self._tree.endUpdateBatch()

