import re
import string

class AutoInsert:

    """Determine how to insert a block of one or more statements into
    a program.  The statement can rely on one or more modules being
    present, which will be imported as necessary.  This is a generic
    implementation.  Each language should subclass this.

    The result is a list of KoInsertion objects, describing how the
    text should be inserted.  This is done to preserve the
    independance of the logic that determines where to insert code
    from the particular format in which the text is stored, and the
    API by which it is manipluated.

    WRT imports, several cases need to be considered.

    1. Import already exists above desired insertion line: don't
       change importing at all.

    2. Import already exists, but below desired insertion line: move
       the import to a reasonable location above insertion line.

    When an import must be inserted there are several cases that must
    be considered to determine where to insert the import.

    1. Only comments and whitespace preceed the desired insertion line:
       insert the import statement after the last comment before the
       desired insertion line.

    2. There are no import statements before the desired insertion
       line, but there is regular code: insert the import statement
       after the last comment before the first non-blank, non-comment
       line.

    3. There are import statements before the desired insertion line:
       Insert the import statement after the last top-level import
       statement before the desired insertion line.
    
    """

    # koIAutoInsertLanguageService interface

    def autoInsert(self, lines, lineno, block, modules):

        """Determine how to insert a block of one or more statements
        into a program.  The statement can rely on one or more modules
        being present, which will be imported as necessary.  This is a
        generic implementation.  Each language should subclass this.
        
        The result is a list of KoInsertion objects, describing how
        the text should be inserted.  This is done to preserve the
        independance of the logic that determines where to insert code
        from the particular format in which the text is stored, and
        the API by which it is manipluated.

        This is an abstract method: it relies on methods that must be
        defined by concrete subclasses.

        lines - the text into which to insert the text of block

        lineno - the 0 based line number after which the block should
                 be inserted.

        block - a list of strings, each representing a single line to
                be inserted.  The lines should not have trailing
                newlines.

        modules - a list of strings containing the names of modules required
                  to support the code in block

        Subclassing

        A subclass should be created for each language that needs to
        support autoinsertion.

        Subclasses must override the following methods, there should
        be no need to override others (see KoPythonAutoInsert for an
        example):

        getImportedModules(self, lines)

        Return a list of the names of the modules imported by the
        program in lines.

        getImportLineNo(self, lineno, lines)

        Return the line number at which additional import statements
        should be inserted.

        makeImportStatement(self, module)

        Return a valid import satement for module.

        isStartOfIndentedBlock(self, line)

        Return 1 if line begins a block that should be indented, else
        0.
        
        """

        insertions = []

        # determine what module imports are needed
        importedModules = self.getImportedModules(lines[:lineno+1])
        importline=0
        for module in modules:
            if module not in importedModules:
                importline = self.getImportLineNo(lines[:lineno+1]);
                insertion = (importline,
                             self.makeImportStatement(module));
                insertions.append(insertion)

        # we should prevent insertion of the statement before the
        # import block, or force module imports that would come after
        # statement to be moved up.

        # determine how block will be inserted

        indent = self.getIndentation(lines, lineno)
        indentedBlock = self.indentBlock(block, indent)
        if lineno > importline+1:
            insertion = (lineno-1, indentedBlock)
        else:
            insertion = (lineno, indentedBlock)
        insertions.append(insertion)

        return insertions


    # helpers
    
    def indentBlock(self, block, indent):
        """Indent each line in block by prepending indent"""

        return map(lambda l, i = indent: i + l, block)
    
    def getIndentation(self, lines, lineno):
        indent = ""
        if lineno < len(lines):
            indent = self.getLineIndentation(lines[lineno])
            if lineno + 1 < len(lines):
                nextLineIndent = self.getLineIndentation(lines[lineno + 1])
                if len(self.untabify(indent)) < \
                   len(self.untabify(nextLineIndent)):
                    indent = nextLineIndent
                elif self.isStartOfIndentedBlock(lines[lineno]):
                    indent = indent + (" " * 4)
        return indent

    def getLineIndentation(self, line):
        """Return the leading whitespace in line"""
        indent = ""
        whitespaceRe = re.compile(r"[ \t]*")
        match = whitespaceRe.match(line)
        if match:
            indent = match.group()
        return indent

    def untabify(self, s):
        """Return s with tabs converted to 8 spaces"""
        return "".join(map(self.untabifyChar, list(s)))
    
    def untabifyChar(self, ch):
        if ch == "\t":
            return " " * 8
        else:
            return ch

    def getImportLineNo(self, lines):
        linenos = []
        for i in range(len(lines)):
            if self.isImportLine(lines[i]):
                linenos.append(i)

        if linenos:
            return linenos[-1]

        # find the last comment line before the first non-comment
        # non-blank line.
        lastCommentLine = 0
        for i in range(len(lines)):
            if self.isCommentLine(lines[i]):
                lastCommentLine = i
            elif lines[i].strip():
                break
        return lastCommentLine

class PythonAutoInsert(AutoInsert):

    def getImportedModules(self, lines):

        """Return a list of the names of modules imported at the top
        level of this file.  Ignore from statements and import as
        statements.

        """

        text = "\n".join(lines)
        importRe = re.compile(r"^import\s+(\w+(?:,\s*\w+)*)", re.MULTILINE)
        matches = importRe.findall(text)
        modules = []
        for match in matches:
            modules += map(string.strip, match.split(","))
        return modules

    def isImportLine(self, line):
        return re.match(r"import\s+\w+", line) is not None

    def isCommentLine(self, line):
        return re.match(r"^\s*#", line) is not None

    def makeImportStatement(self, module):
        return ["import %s" % module]

    def isStartOfIndentedBlock(self, line):
        return re.match(r".*:\s*$", line) is not None


class PerlAutoInsert(AutoInsert):

    def getImportedModules(self, lines):
        text = "\n".join(lines)
        importRe = re.compile(r"^use\s+([\w:]+(?:,\s*\w+)*);", re.MULTILINE)
        matches = importRe.findall(text)
        modules = []
        for match in matches:
            modules += map(string.strip, match.split(","))
        return modules

    def isImportLine(self, line):
        return re.match(r"use\s+\w+;", line) is not None

    def isCommentLine(self, line):
        return re.match(r"^\s*#", line) is not None

    def makeImportStatement(self, module):
        return ["use %s;" % module]

    def isStartOfIndentedBlock(self, line):
        return re.match(r".*\{\s*", line) is not None



class PHPAutoInsert(AutoInsert):
    # regex copied from koPHPCompletion.py
    re_include = re.compile(r'\s*(?:require|include|require_once|include_once)\s*[\(|\s]*[\"|\']\s*(.*?)\s*[\"|\'][\)|\s]*;',re.I|re.M)

    def getImportedModules(self, lines):
        text = "\n".join(lines)
        matches = self.re_include.findall(text)
        modules = []
        for match in matches:
            modules += map(string.strip, match.split(","))
        return modules

    def isImportLine(self, line):
        #return self.re_include.match(line) is not None
        # always insert after first open tag that is not on a line by itself.
        #XXX not a perfect world here for php.
        if not re.match(r'\?>', line): 
            return re.match(r'(<\?php)', line) is not None
        return 0
    
    def isCommentLine(self, line):
        return re.match(r"^\s*(?:#|//)", line) is not None

    def makeImportStatement(self, module):
        return ["require_once(\"%s\");" % module]

    def isStartOfIndentedBlock(self, line):
        return re.match(r".*\{\s*", line) is not None

