# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python XPCOM language bindings.
#
# The Initial Developer of the Original Code is
# Activestate Tool Corp.
# Portions created by the Initial Developer are Copyright (C) 2000
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Mark Hammond <MarkH@activestate.com>
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
#

# The XPCOM (Cross Platform COM) package.
import exceptions


import sys
#sys.stdout = open("pystdout.log", "w")
if sys.version_info >= (2, 3):
    # sick off the new hex() warnings, and no time to digest what the 
    # impact will be!
    import warnings
    warnings.filterwarnings("ignore", category=FutureWarning, append=1)

import os
_location = os.path.dirname(__file__)

# A global "verbose" flag - currently used by the
# server package to print trace messages
verbose = 0
hr_map = {}

class nsStackFrame:
    language = 0 # what should it be?
    languageName = "Python"

    def __init__(self, tb):
        try:
            # magic to hook up com interface after xpcom is loaded
            if not hasattr(nsStackFrame, "_com_interfaces_"):
                from xpcom import components
                nsStackFrame._com_interfaces_ = [components.interfaces.nsIStackFrame]
        except Exception, e:
            # not initialized yet, cannot make this an xpcom component
            pass

        self.caller = None

        import linecache
    
        f = tb.tb_frame
        self.lineNumber = tb.tb_lineno
        co = f.f_code
        self.filename = co.co_filename
        self.name = co.co_name
        linecache.checkcache(self.filename)
        line = linecache.getline(self.filename, self.lineNumber)
        if line: self.sourceLine = line.strip()
        else: self.sourceLine = None

    def toString(self):
        return 'File "%s", line %d, in %s' % (self.filename,
                                              self.lineNumber,
                                              self.name)

# The standard XPCOM exception object.
# Instances of this class are raised by the XPCOM extension module.
class Exception(exceptions.Exception):
    def __init__(self, errno, message = None):
        exceptions.Exception.__init__(self, errno, message)
        assert int(errno) == errno, "The errno param must be an integer"
        try:
            # magic to hook up com interface after xpcom is loaded
            if not hasattr(Exception, "_com_interfaces_"):
                import components
                Exception._com_interfaces_ = [components.interfaces.nsIException]
        except exceptions.Exception, e:
            # not initialized yet, cannot make this an xpcom component
            pass

        if not hr_map:
            import nsError
            for name, val in nsError.__dict__.items():
                if type(val)==type(0):
                    hr_map[val] = name

        self.errno = errno
        self.message = message
        self.result = errno
        self.name = hr_map.get(self.errno, "")
        self.filename = None
        self.lineNumber = 0
        self.columnNumber = 0
        self.location = None
        self.inner = None
        self.data = None

    def setExcInfo(self, exc_info):
        if exc_info[2]:
            self.setStackFrame(exc_info[2])
        
    def setStackFrame(self, tb):
        caller = None
        while tb.tb_next is not None:
            f = nsStackFrame(tb)
            if not f.filename.startswith(_location):
                prev_caller = caller
                caller = f
                caller.caller = prev_caller
            tb = tb.tb_next

        f = nsStackFrame(tb)
        if caller and f.filename[0] == "<":
            self.location = caller
            self.location.caller = caller.caller
        else:
            self.location = f
            self.location.caller = caller
        self.lineNumber = f.lineNumber
        self.filename = f.filename
        self.name = f.name

    def setData(self, exc_info):
        import traceback
        textObj = components.classes["@mozilla.org/supports-string;1"]\
                .createInstance(components.interfaces.nsISupportsString)
        textObj.data = '\n'.join(traceback.format_tb(exc_info[2]))
        self.data = textObj;

    def __str__(self):
        if self.message is None:
            return self.name
        return str(self.message)
    def toString(self):
        return str(self)

# An alias for Exception - allows code to say "from xpcom import COMException"
# rather than "Exception" - thereby preventing clashes.
COMException = Exception

# Exceptions thrown by servers.  It can be good for diagnostics to
# differentiate between a ServerException (which was presumably explicitly thrown)
# and a normal exception which may simply be propagating down.
# (When ServerException objects are thrown across the XPConnect
# gateway they will be converted back to normal client exceptions if
# subsequently re-caught by Python
class ServerException(Exception):
    def __init__(self, errno=None, *args, **kw):
        if errno is None:
            import nsError
            errno = nsError.NS_ERROR_FAILURE
        Exception.__init__(self, errno, *args, **kw)

# Some global functions.
